// test plugin

#include "vrayplugins.h"
#include "vrayinterface.h"
#include "vrayrenderer.h"
#include "brdfs.h"
#include "vraytexutils.h"
// include "basetexture.h"
#include "defparams.h"

namespace VUtils{
class BRDFNormalTestSampler: public BRDFSampler, public BSDFSampler {
//protected:

    int orig_backside;
    Vector normal, gnormal;
    Matrix nm, inm;

    // Assigned in init()
    Color m_diffuse_color = Color{1.f, 1.f, 1.f};

public:
    void init(const VR::VRayContext &rc);

    // From BRDFSampler
    Vector getDiffuseNormal(const VR::VRayContext &rc);
    Color getDiffuseColor(Color &lightColor);
    Color getLightMult(Color &lightColor);
    Color getTransparency(const VR::VRayContext &rc);

    Color eval(const VR::VRayContext &rc, const Vector &direction, Color &lightColor, Color &origLightColor, float probLight, int flags);
    void traceForward(VR::VRayContext &rc, int doDiffuse);

    int getNumSamples(const VR::VRayContext &rc, int doDiffuse);
    VR::VRayContext* getNewContext(const VR::VRayContext &rc, int &samplerID, int doDiffuse);
    VR::ValidType setupContext(const VR::VRayContext &rc, VRayContext &nrcm, float uc, int doDiffuse);

    BRDFSampler* getBRDF(BSDFSide side);
    
    RenderChannelsInfo* getRenderChannels(void);

};

}

using namespace VR;

struct BRDFNormalTestParams: VRayParameterListDesc {
    BRDFNormalTestParams(void) {
        addParamTexture("diffuse_color", Color(0.8f, 0.8f, 0.8f), -1, "Diffuse color.");
    }
};

struct BRDFNormalTest: VRayBSDF {
    BRDFNormalTest(VRayPluginDesc *pluginDesc):VRayBSDF(pluginDesc) {
        paramList->setParamCache("diffuse_color", &m_diffuse_color);
    }

    // From VRayBSDF
    void renderBegin(VRayRenderer *vray) {return;};
    void renderEnd(VRayRenderer *vray) {return;};
    void frameBegin(VRayRenderer *vray);
    void frameEnd(VRayRenderer *vray);

    // From BRDFInterface
    BSDFSampler *newBSDF(const VRayContext &rc, BSDFFlags flags);
    void deleteBSDF(const VRayContext &rc, BSDFSampler *bsdf);

    virtual PluginInterface* newInterface(InterfaceID id) {
        if (id==EXT_BSDF) return static_cast<PluginInterface*>((BSDFInterface*) this);
        return VRayPlugin::newInterface(id);
    }

    int isOpaque(void) VRAY_OVERRIDE {
        return getBRDFPluginFlags(baseBRDFParam);
    }

private:
    BRDFPool<BRDFNormalTestSampler> pool;
    Color m_diffuse_color = Color{1.f, 1.f, 1.f};

    // Parameters
    PluginBase *baseBRDFParam;
    PluginBase *bumpTexPlugin;
};

void BRDFNormalTest::frameBegin(VRayRenderer *vray) {
    // Make new VRayContext
    VRayThreadData* vr_tdata = vray->getThreadData(0);
    VRayCore* vcore = static_cast<VRayCore *>(vray);
    VRayContext rc;
    rc.init(vr_tdata, vcore);
    
    //VRayPluginParameter* diffuse_color = this->getParameter("diffuse_color");

    pool.init(vray->getSequenceData().maxRenderThreads);
    return;
}

void BRDFNormalTest::frameEnd(VRayRenderer *vray) {
    pool.freeMem();
    return;
}

BSDFSampler* BRDFNormalTest::newBSDF(const VRayContext &rc, BSDFFlags flags) {
    BRDFNormalTestSampler *bsdf = pool.newBRDF(rc);
    if (!bsdf) return NULL;
    bsdf->init(rc);
    return bsdf;
}

void BRDFNormalTest::deleteBSDF(const VRayContext &rc, BSDFSampler *bsdf) {
    if (!bsdf) return;
    BRDFNormalTestSampler *nbsdf = static_cast<BRDFNormalTestSampler*>(bsdf);
    pool.deleteBRDF(rc, nbsdf);
}

using namespace VUtils;

// Implementation of Sampler.
void BRDFNormalTestSampler::init(const VR::VRayContext &rc) {
    orig_backside = rc.rayresult.realBack;

    normal = rc.rayresult.normal;    //smoothed normal
    gnormal = rc.rayresult.gnormal;  //geometric normal
    if (rc.rayresult.realBack) {
        normal = -normal;
        gnormal = -gnormal;
    }
    
    // Compute normal matrix, generates matrix for transforming from normal
    // space to world coordinates.
    if (rc.rayparams.viewDir*normal>0.0f) {
        makeNormalMatrix(gnormal, nm);
    } else {
        makeNormalMatrix(normal, nm);
    }
    inm = inverse(nm);

    m_diffuse_color.r = 1.f;
    m_diffuse_color.g = 1.f;
    m_diffuse_color.b = 1.f;

    return;
}

// FROM BRDFSampler

// Returns normal to use for diffuse lightinh
Vector BRDFNormalTestSampler::getDiffuseNormal(const VR::VRayContext &rc) {
    return rc.rayresult.origNormal;
}

// Returns amount of light reflected diffusely by the surface.
Color BRDFNormalTestSampler::getDiffuseColor(Color &lightColor) {
    Color ret = m_diffuse_color*lightColor;

    ret = Color{1.f, 1.f, 1.f};
    lightColor.makeZero();
    return ret;
}

// Total amount of light reflected by the surface.
Color BRDFNormalTestSampler::getLightMult(Color &lightColor) {
    float s = 0.f;
    Color ret = (m_diffuse_color + Color(s,s,s)) * lightColor;
    
    //lightColor is amount of light transmitted through the BRDF.
    lightColor.makeZero();
    
    ret = Color{1.f, 1.f, 1.f};
    return ret;
}

// Returns transparency of the BRDF at the given point.
Color BRDFNormalTestSampler::getTransparency(const VR::VRayContext &rc) {
	return Color(0.f,0.f,0.f);
}

// Returns the amount of light transmitted from the given direction into the viewing direction.
Color BRDFNormalTestSampler::eval(const VR::VRayContext &rc, const Vector &direction,
    Color &lightColor, Color &origLightColor, float probLight,
    int flags) {
    Color res;
    float k;
    // If diffuse component is not required, return black
    if (0==(flags & FBRDF_DIFFUSE)) 
    {
        res.makeZero();
    } 
    else 
    {//rayparams.tracedRay
        float cs = direction * rc.rayresult.normal;
        cs = cs < 0.0f ? 0.0f : cs;
    
        m_diffuse_color.r = rc.rayresult.normal.x;
        m_diffuse_color.g = rc.rayresult.normal.y;
        m_diffuse_color.b = rc.rayresult.normal.z;
        
        k = cs;
        if (rc.rayparams.localRayType & RT_IS_GATHERING_POINT) {
            float probReflection=k;
            k = cs*getReflectionWeight(probLight, probReflection);
        }
        res = m_diffuse_color * k *lightColor;
    }

    lightColor.makeZero();
    origLightColor.makeZero();

    return res;
}

void BRDFNormalTestSampler::traceForward(VR::VRayContext &rc, int doDiffuse) {
	BRDFSampler::traceForward(rc, doDiffuse);
	if (doDiffuse) rc.mtlresult.color+=rc.evalDiffuse()*m_diffuse_color;
    Fragment *f=rc.mtlresult.fragment;
    if(f){
        VUtils::Color raw_gi=rc.evalDiffuse();
        VUtils::Color diffuse_gi=raw_gi*m_diffuse_color;
        f->setChannelDataByAlias(REG_CHAN_VFB_RAWGI,&raw_gi);
        f->setChannelDataByAlias(REG_CHAN_VFB_RAWTOTALLIGHT,&raw_gi);
        f->setChannelDataByAlias(REG_CHAN_VFB_GI,&diffuse_gi);
        f->setChannelDataByAlias(REG_CHAN_VFB_TOTALLIGHT,&diffuse_gi);
        f->setChannelDataByAlias(REG_CHAN_VFB_DIFFUSE,&raw_gi);
    }
}

// Returns the number of desired samples for evaluating the BRDF.
int BRDFNormalTestSampler::getNumSamples(const VR::VRayContext &rc, int doDiffuse) {
    if(rc.rayparams.currentPass==RPASS_GI ||
        (rc.rayparams.rayType & RT_LIGHT)!=0) {
        return 0;
    }
    return 8;
}

// Returns a new ray context for integrating the BRDF.
VR::VRayContext* BRDFNormalTestSampler::getNewContext(const VR::VRayContext &rc, int &samplerID, int doDiffuse) {
	if (2==doDiffuse) return NULL;
    //TODO use specular strength to weight the samples.
    return NULL;
}

// This sets up the given context for tracing a ray for BRDF integration.
ValidType BRDFNormalTestSampler::setupContext(const VR::VRayContext &rc, VR::VRayContext &nrc, float uc, int doDiffuse) {
    Vector dir = getDiffuseDir(uc, BRDFSampler::getDMCParam(nrc,1),nrc.rayparams.rayProbability);
    if (dir*gnormal<0.0f) return false;
    VR::getReflectDerivs(rc.rayparams.viewDir, dir, rc.rayparams.dDdx, rc.rayparams.dDdy, nrc.rayparams.dDdx, nrc.rayparams.dDdy);
    nrc.rayparams.tracedRay.dir=nrc.rayparams.viewDir=dir;
    return true;
}

RenderChannelsInfo* BRDFNormalTestSampler::getRenderChannels(void) { return &RenderChannelsInfo::reflectChannels; }


BRDFSampler* BRDFNormalTestSampler::getBRDF(BSDFSide side) {
    if (side==bsdfSide_front) {
        if (orig_backside) return NULL;
    } else {
        if (!orig_backside) return NULL;
    }
    return static_cast<BRDFSampler*>(this);
}


#define BRDFTest_PluginID PluginID(LARGE_CONST(2017091301))
// #ifdef __VRAY40__
// SIMPLE_PLUGIN_LIBRARY(BRDFTest_PluginID, "BRDFTest", "BRDFTestLibText", "a simple test plugin/lib", BRDFTest, BRDFTest_Params, EXT_BSDF);
// #else
SIMPLE_PLUGIN_LIBRARY(BRDFTest_PluginID, EXT_BSDF, "BRDFNormalTest", "BRDFTestLibText", BRDFNormalTest, BRDFNormalTestParams);
// #endif

//PLUGIN_DESC(BRDFTest_PluginID, EXT_BSDF, "BRDFTest", BRDFTest, BRDFTest_Params);

